import json
import jsonref
import inflection
from pydantic import BaseModel
from typing import List
from .action import Action
from .trigger import Trigger
    
def add_tool_name(tool_name: str, action_name: str) -> str:
    return f"{tool_name.lower()}_{inflection.underscore(action_name)}"

class Tool(BaseModel):
    _custom_id: str = "default"  # Add an internal variable to hold the custom repo name

    @property
    def custom_id(self) -> str:
        return self._custom_id
    
    @custom_id.setter
    def custom_id(self, value: str):
        self._custom_id = value  # Set the internal variable

    def actions(self) -> List[Action]:
        raise NotImplementedError("This method should be overridden by subclasses.")
    
    def triggers(self) -> List[Trigger]:
        raise NotImplementedError("This method should be overridden by subclasses.")

    def execute_action(self, action_name: str, request_data: dict, authorisation_data: dict) -> dict:
        tool_name = self.__class__.__name__
        for action in self.actions():
            action_name_shared = add_tool_name(tool_name, action.__name__)
            if action_name_shared == action_name:
                print(f"Executing {action.__name__} on Tool: {tool_name} with request data {request_data} and authorisation data {authorisation_data}")
                try:
                    request_schema = action().request_schema
                    req = request_schema.model_validate_json(json_data=json.dumps(request_data))
                    return action().execute(req, authorisation_data)
                except Exception as e:
                    return {"status": "failure", "details": str(e)}
        return {"status": "failure", "details": "Action not found"}
    
    def transform_trigger_payload(self, payload: dict) -> dict:
        tool_name = self.__class__.__name__
        for trigger in self.triggers():
            try:
                (match, converted_payload) = trigger().check_and_convert_to_identifier_payload_schema(payload)
                if match:
                    converted_payload["trigger_name"] = add_tool_name(tool_name, trigger.__name__)
                    return converted_payload
            except Exception as e:
                print(f"Error transforming trigger payload for Tool: {tool_name} with payload {payload}: {str(e)}")
        return {"status": "failure", "details": "Trigger not found"}
    
    def set_webhook_url(self, trigger_name: str, authorisation_data: dict, trigger_config: dict) -> dict: 
        tool_name = self.__class__.__name__
        for trigger in self.triggers():
            trigger_name_shared = add_tool_name(tool_name, trigger.__name__)
            if trigger_name_shared == trigger_name:
                print(f"Setting webhook URL for Trigger: {trigger_name} on Tool: {tool_name} with URL {trigger_config['webhook_url']} and authorisation data {authorisation_data}")
                try:
                    response = trigger().set_webhook_url(authorisation_data, trigger_config)
                    return {"status": "success", "connection_data": response}
                except Exception as e:
                    return {"status": "failure", "details": str(e)}
        return {"status": "failure", "details": "Trigger not found"}
        
        
    def json_schema(self):
        tool_name = self.__class__.__name__
        return {
            "Name": tool_name,
            "Description": self.__doc__.strip() if self.__doc__ else None,
            "CustomID": self.custom_id,
            "Actions": [
                {
                    "name": add_tool_name(tool_name, action.__name__),
                    "display_name": action().display_name,
                    "description": action.__doc__.strip() if action.__doc__ else None,
                    "parameters": jsonref.loads(json.dumps(action().request_schema.model_json_schema())),
                    "response": jsonref.loads(json.dumps(action().response_schema.model_json_schema()))
                } for action in self.actions()
            ],
            "Triggers": [
                {
                    "name": add_tool_name(tool_name, trigger.__name__),
                    "display_name": trigger().display_name,
                    "description": trigger.__doc__.strip() if trigger.__doc__ else None,
                    "payload": jsonref.loads(json.dumps(trigger().payload_schema.model_json_schema())),
                    "config": jsonref.loads(json.dumps(trigger().trigger_config.model_json_schema())),
                    "instructions": trigger().trigger_instructions
                } for trigger in self.triggers()
            ],
        }

